{{question_text}}
{% if use_code %}
The options are
{% for option in question_options %}
{{ loop.index0 }}: {{option}}
{% endfor %}
{% else %}
The options are:
{% for option in question_options %}
{{option}}
{% endfor %}
{% endif %}
{% if num_selections %}
You have to include {{num_selections}} options in your answer.
{% endif %}
